from typing import get_origin

from automappy.exceptions import ValueTypeError
from automappy.mappers.registry import MapperRegistry
from automappy.types.type_pair import TypePair


class Mapper[F, T]:
    def __init__(self, from_: type[F], to: type[T]):
        self._tp = TypePair(from_, to)

    @classmethod
    def create_mapper[F_, T_](cls, from_: type[F_], to: type[T_]) -> "Mapper[F_, T_]":
        return Mapper(from_, to)

    def map(self, value: F) -> T:
        self._assert_is_type(value)

        mapper = MapperRegistry.find_mapper(self._tp)
        return mapper.map(self._tp, value)

    def _assert_is_type(self, value: F) -> None:
        expected_type = get_origin(self._tp.from_) or self._tp.from_

        if type(value) is not expected_type:
            raise ValueTypeError(value, self._tp.from_)



from automappy.exceptions import ObjectMapperNotFoundError
from automappy.mappers.collection_mapper import CollectionMapper
from automappy.mappers.mapping_mapper import MappingMapper
from automappy.mappers.object_mapper import ObjectMapper
from automappy.mappers.primitive_mapper import PrimitiveMapper
from automappy.types.type_pair import TypePair


class MapperRegistry:
    _mappers: list[ObjectMapper] = [
        PrimitiveMapper(),
        CollectionMapper(),
        MappingMapper(),
    ]

    @classmethod
    def find_mapper(cls, tp: TypePair) -> ObjectMapper:
        for mapper in cls._mappers:
            if mapper.is_compatible(tp):
                return mapper

        raise ObjectMapperNotFoundError(tp)



from typing import Protocol

from automappy.types.type_pair import TypePair


class ObjectMapper(Protocol):
    def is_compatible(self, tp: TypePair) -> bool: ...

    def map[F, T](self, tp: TypePair[F, T], from_: F) -> T: ...



from typing import NamedTuple
import pytest
from automappy.exceptions import ObjectMapperNotFoundError, ValueTypeError
from automappy.mapper import Mapper


def test_incorrect_map_type() -> None:
    m = Mapper.create_mapper(int, str)

    with pytest.raises(ValueTypeError):
        m.map("foo")  # type: ignore[reportArgumentType]


def test_simple() -> None:
    m = Mapper.create_mapper(int, str)

    assert m.map(42) == "42"
    assert m.map(0) == "0"
    assert m.map(-42) == "-42"

    m = Mapper.create_mapper(str, int)
    assert m.map("42") == 42
    assert m.map("0") == 0
    assert m.map("-42") == -42

    m = Mapper.create_mapper(float, str)
    assert m.map(42.0) == "42.0"
    assert m.map(0.0) == "0.0"
    assert m.map(-42.0) == "-42.0"

    m = Mapper.create_mapper(str, float)
    assert m.map("42.0") == 42.0
    assert m.map("0.0") == 0.0
    assert m.map("-42.0") == -42.0


def test_object_mapper_not_found() -> None:
    m = Mapper.create_mapper(int, NamedTuple)

    with pytest.raises(ObjectMapperNotFoundError):
        m.map(42)


def test_list_to_set() -> None:
    m = Mapper.create_mapper(list[int], set[int])

    assert m.map([1, 2, 3]) == {1, 2, 3}
    assert m.map([]) == set()
    assert m.map([42]) == {42}
    assert m.map([1, 2, 3, 1, 2, 3]) == {1, 2, 3}


def test_list_content() -> None:
    m = Mapper.create_mapper(list[int], list[str])

    assert m.map([1, 2, 3]) == ["1", "2", "3"]
    assert m.map([]) == []
    assert m.map([42]) == ["42"]


def test_list_content_mapper_not_found() -> None:
    m = Mapper.create_mapper(list[int], list[NamedTuple])

    with pytest.raises(ObjectMapperNotFoundError):
        m.map([1, 2, 3])


def test_dict_mapper() -> None:
    m = Mapper.create_mapper(dict[int, str], dict[str, int])

    assert m.map({1: "1", 2: "2", 3: "3"}) == {"1": 1, "2": 2, "3": 3}
    assert m.map({}) == {}
    assert m.map({42: "42"}) == {"42": 42}


def test_dict_list_mapper() -> None:
    m = Mapper.create_mapper(dict[int, list[int]], dict[str, set[int]])

    assert m.map({1: [1, 2, 2, 3], 2: [4, 5, 6], 3: [7, 8, 9]}) == {
        "1": {1, 2, 3},
        "2": {4, 5, 6},
        "3": {7, 8, 9},
    }
    assert m.map({}) == {}
    assert m.map({42: [42]}) == {"42": {42}}



from automappy.mappers.object_mapper import ObjectMapper
from automappy.types.type_pair import TypePair


class PrimitiveMapper(ObjectMapper):
    def is_compatible(self, tp: TypePair) -> bool:
        return self._is_primitive(tp.from_) and self._is_primitive(tp.to)

    def map[F, T](self, tp: TypePair[F, T], from_: F) -> T:
        return tp.to.__call__(from_)

    def _is_primitive(self, type_: type) -> bool:
        return type_ in {int, float, str, bool}



from typing import NamedTuple


class TypePair[F, T](NamedTuple):
    from_: type[F]
    to: type[T]

    def __repr__(self) -> str:
        return f"{self.from_} -> {self.to}"

    def __str__(self) -> str:
        return self.__repr__()



from collections.abc import Collection
from typing import cast
from automappy.mappers.object_mapper import ObjectMapper
from automappy.types.type_pair import TypePair


class CollectionMapper(ObjectMapper):
    def is_compatible(self, tp: TypePair) -> bool:
        return self._is_collection(tp.from_) and self._is_collection(tp.to)

    def map[F, T](self, tp: TypePair[F, T], from_: F) -> T:
        from automappy.mappers.registry import MapperRegistry

        assert self._is_collection(type(from_)), "from_ must be a collection"

        from_content_type = self._get_collection_content_type(tp.from_)
        to_content_type = self._get_collection_content_type(tp.to)

        content_tp = TypePair(from_content_type, to_content_type)
        content_mapper = MapperRegistry.find_mapper(content_tp)

        return tp.to.__call__(
            [content_mapper.map(content_tp, value) for value in cast(Collection, from_)]
        )

    def _is_collection(self, type_: type) -> bool:
        return hasattr(type_, "__iter__") and not self._is_mapping(type_)

    def _is_mapping(self, type_: type) -> bool:
        return (
            hasattr(type_, "items")
            and hasattr(type_, "keys")
            and hasattr(type_, "values")
        )

    def _get_collection_content_type(self, type_: type) -> type:
        return type_.__args__[0]



from collections.abc import Mapping
from typing import cast
from automappy.mappers.object_mapper import ObjectMapper
from automappy.types.type_pair import TypePair


class MappingMapper(ObjectMapper):
    def is_compatible(self, tp: TypePair) -> bool:
        return self._is_mapping(tp.from_) and self._is_mapping(tp.to)

    def map[F, T](self, tp: TypePair[F, T], from_: F) -> T:
        from automappy.mappers.registry import MapperRegistry

        assert self._is_mapping(type(from_)), "from_ must be a mapping"

        from_key_type = self._get_mapping_key_type(tp.from_)
        to_key_type = self._get_mapping_key_type(tp.to)
        key_tp = TypePair(from_key_type, to_key_type)
        key_mapper = MapperRegistry.find_mapper(key_tp)

        from_value_type = self._get_mapping_value_type(tp.from_)
        to_value_type = self._get_mapping_value_type(tp.to)
        value_tp = TypePair(from_value_type, to_value_type)
        value_mapper = MapperRegistry.find_mapper(value_tp)

        return tp.to.__call__(
            {
                key_mapper.map(key_tp, key): value_mapper.map(value_tp, value)
                for key, value in cast(Mapping, from_).items()
            }
        )

    def _get_mapping_key_type(self, type_: type) -> type:
        return type_.__args__[0]

    def _get_mapping_value_type(self, type_: type) -> type:
        return type_.__args__[1]

    def _is_mapping(self, type_: type) -> bool:
        return (
            hasattr(type_, "items")
            and hasattr(type_, "keys")
            and hasattr(type_, "values")
        )



from typing import Any

from automappy.types.type_pair import TypePair


class ValueTypeError(TypeError):
    def __init__(self, value: Any, expected_type: type):
        self.value = value
        self.expected_type = expected_type
        super().__init__(f"Value '{value}' is not of type '{expected_type}'")


class ObjectMapperNotFoundError(ValueError):
    def __init__(self, tp: TypePair):
        self.tp = tp
        super().__init__(f"Object mapper not found for type pair '{tp}'")
